use Link;
use searchers::Searcher;

use reqwest;

use serde_json;
use serde_json::Value;

use retry;
use retry::retry;
use retry::delay::Fixed;

use std;
use std::io::Read;

//use std::any::Any;
use mopa::Any;

const NUM_RETRIES: usize = 10;
const RETRY_WAIT_MILLIS: u64 = 500;

#[derive(Debug)]
pub struct YugiohCard {
    pub name: String,
    pub card_type: String,
    pub text: String,
    pub subtype: Option<String>,
    pub family: Option<String>,
    pub atk: Option<u64>,
    pub def: Option<u64>,
    pub level: Option<u64>,
    pub url: String,
    pub image_url: String
}

impl Link for YugiohCard {
    fn label (&self) -> &str {
        &self.name
    }

    fn url (&self) -> &str {
        &self.url
    }

    fn as_any (&self) -> &Any {
        self
    }
}

pub struct YugiohSearcher {}

impl YugiohSearcher {
    pub fn new () -> YugiohSearcher {
        YugiohSearcher {}
    }

    fn do_search (&self, name: &str) -> Result<String, Error> {
        let mut contents = String::new();
        let api_url = &format!("http://yugiohprices.com/api/card_data/{}", name);
        let mut response = retry(Fixed::from_millis(RETRY_WAIT_MILLIS).take(NUM_RETRIES), || {
            reqwest::get(api_url)
        })?;
        response.read_to_string(&mut contents)?;
        Result::Ok(contents)
    }
}

fn parse_entry (page: String) -> Result<YugiohCard, Error> {
    let parsed: Value = serde_json::from_str(&page)?;
    let ref parsed_entry = parsed["data"];
    if let Some(_) = parsed_entry.as_object() {
        let card_name = parsed_entry["name"].as_str().map(String::from).expect("expected name in json data");
        let card_url = format!("https://yugipedia.com/wiki/{}", card_name.replace(" ", "_"));
        let card_image_url = format!("http://static.api3.studiobebop.net/ygo_data/card_images/{}.jpg", card_name.replace(" ", "_").replace("-", "_"));
        Result::Ok(YugiohCard {
            name: card_name,
            card_type: parsed_entry["card_type"].as_str().map(String::from).expect("expected card type in json data"),
            text: parsed_entry["text"].as_str().map(String::from).expect("expected text in json data"),
            subtype: parsed_entry["type"].as_str().map(String::from),
            family: parsed_entry["family"].as_str().map(String::from),
            atk: parsed_entry["atk"].as_u64(),
            def: parsed_entry["def"].as_u64(),
            level: parsed_entry["level"].as_u64(),
            url: card_url,
            image_url: card_image_url
        })
    } else {
        Result::Err(Error::Other(String::from("No card info found")))
    }
}

impl Searcher<YugiohCard> for YugiohSearcher {
    fn exact_search (&self, name: &str) -> Option<YugiohCard> {
        self.do_search(name).and_then(parse_entry).ok()
    }
}

#[derive(Debug)]
enum Error {
    Http(reqwest::Error),
    Io(std::io::Error),
    Json(serde_json::Error),
    Other(String)
}

impl From<reqwest::Error> for Error {
    fn from (error: reqwest::Error) -> Error {
        Error::Http(error)
    }
}

impl From<std::io::Error> for Error {
    fn from (error: std::io::Error) -> Error {
        Error::Io(error)
    }
}

impl From<serde_json::Error> for Error {
    fn from (error: serde_json::Error) -> Error {
        Error::Json(error)
    }
}

impl From<retry::Error<reqwest::Error>> for Error {
    fn from (err: retry::Error<reqwest::Error>) -> Error {
        match err {
            retry::Error::Operation { error, total_delay, tries } => {
                Error::Http(error)
            }
        }
    }
}
